from collections import namedtuple
from enum import Enum, unique

from microcosm_flask.formatting import (
    CSVFormatter,
    JSONFormatter,
    HTMLFormatter,
    TextFormatter,
)

ResponseFormatSpec = namedtuple("ResponseFormatSpec", ["content_type", "formatter", "priority"])


@unique
class ResponseFormats(Enum):
    CSV = ResponseFormatSpec(
        content_type=CSVFormatter.CONTENT_TYPE,
        formatter=CSVFormatter,
        priority=100,
    )
    JSON = ResponseFormatSpec(
        content_type=JSONFormatter.CONTENT_TYPE,
        formatter=JSONFormatter,
        priority=1,
    )
    HTML = ResponseFormatSpec(
        content_type=HTMLFormatter.CONTENT_TYPE,
        formatter=HTMLFormatter,
        priority=10,
    )
    TEXT = ResponseFormatSpec(
        content_type=TextFormatter.CONTENT_TYPE,
        formatter=TextFormatter,
        priority=150,
    )

    @property
    def content_type(self):
        return self.value.content_type

    @property
    def priority(self):
        return self.value.priority

    def matches(self, content_types):
        for content_type in content_types.split(","):
            if self.matches_content_type(content_type):
                return True
        return False

    def matches_content_type(self, content_type):
        this = self.content_type.split("/", 1)
        that = content_type.split(";", 1)[0].split("/", 1)

        for (this_part, that_part) in zip(this, that):
            if that_part != "*" and this_part != that_part:
                return False

        return True

    @classmethod
    def prioritized(cls):
        return sorted(cls, key=lambda this: this.priority)



"""
Response formatting base class.

"""
from abc import ABCMeta, abstractmethod
from binascii import hexlify
try:
    import spooky
except ImportError:
    spooky = None

from flask import Response
from werkzeug.http import quote_etag
from werkzeug.utils import get_content_type


class BaseFormatter(metaclass=ABCMeta):

    def __init__(self, response_schema=None):
        # Formatting could need the response schema
        # e.g. to specify column ordering in CSV response
        self.response_schema = response_schema

    def __call__(self, response_data, headers=None, **kwargs):
        response = self.build_response(response_data)
        headers = self.build_headers(headers=headers or {}, **kwargs)
        response.headers.extend(headers)
        self.build_etag(response, **kwargs)
        return response

    @property
    @abstractmethod
    def content_type(self):
        pass

    def format(self, response_data):
        return response_data

    def build_response(self, response_data):
        return Response(
            self.format(response_data),
            content_type=get_content_type(self.content_type, Response.charset)
        )

    def build_headers(self, headers, **kwargs):
        return headers

    def build_etag(self, response, include_etag=True, **kwargs):
        """
        Add an etag to the response body.

        Uses spooky where possible because it is empirically fast and well-regarded.

        See: http://blog.reverberate.org/2012/01/state-of-hash-functions-2012.html

        """
        if not include_etag:
            return

        if not spooky:
            # use built-in md5
            response.add_etag()
            return

        # use spooky
        response.headers["ETag"] = quote_etag(
            hexlify(
                spooky.hash128(
                    response.get_data(),
                ).to_bytes(16, "little"),
            ).decode("utf-8"),
        )



"""
Generate JSON Schema for Marshmallow schemas.

"""
from logging import getLogger

from marshmallow import fields

from microcosm_flask.fields import (
    EnumField,
    LanguageField,
    QueryStringList,
    TimestampField,
    URIField,
)
from microcosm_flask.naming import name_for
from microcosm_flask.swagger.naming import type_name


logger = getLogger("microcosm_flask.swagger")


# see: https://github.com/marshmallow-code/apispec/blob/dev/apispec/ext/marshmallow/swagger.py
FIELD_MAPPINGS = {
    EnumField: (None, None),
    LanguageField: ("string", "language"),
    QueryStringList: ("array", None),
    URIField: ("string", "uri"),
    fields.Boolean: ("boolean", None),
    fields.Date: ("string", "date"),
    fields.DateTime: ("string", "date-time"),
    fields.Decimal: ("number", None),
    fields.Dict: ("object", None),
    fields.Email: ("string", "email"),
    fields.Field: ("object", None),
    fields.Float: ("number", "float"),
    fields.Integer: ("integer", "int32"),
    fields.List: ("array", None),
    fields.Method: ("object", None),
    fields.Nested: (None, None),
    fields.Number: ("number", None),
    fields.Raw: ("object", None),
    fields.String: ("string", None),
    fields.Time: ("string", None),
    fields.URL: ("string", "url"),
    fields.UUID: ("string", "uuid"),
}


SWAGGER_TYPE = "__swagger_type__"
SWAGGER_FORMAT = "__swagger_format__"


def is_int(value):
    try:
        int(value)
    except Exception:
        return False
    else:
        return True


def swagger_field(field, swagger_type="string", swagger_format=None):
    """
    Decorator for an existing field type to coerce a specific swagger type/format.

    Usage:

        class MySchema(Schema):
             myfield = swagger_field(fields.Foo(), swagger_type="string")

    """
    setattr(field, SWAGGER_TYPE, swagger_type)
    setattr(field, SWAGGER_FORMAT, swagger_format)
    return field


def resolve_tagged_field(field):
    """
    Fields tagged with `swagger_field` shoudl respect user definitions.

    """
    field_type = getattr(field, SWAGGER_TYPE)
    field_format = getattr(field, SWAGGER_FORMAT, None)

    if isinstance(field_type, list):
        # Ideally we'd use oneOf here, but OpenAPI 2.0 uses the 0.4-draft jsonschema
        # which doesn't include oneOf. Upgrading to OpenAPI 3.0 ought to suffice.
        return dict()
    elif field_format:
        return dict(
            type=field_type,
            format=field_format,
        )
    else:
        return dict(
            type=field_type,
        )


def resolve_enum_field(field):
    """
    Enum fields should include choices.

    """
    enum = getattr(field, "enum", None)
    enum_values = [
        choice.value if field.by_value else choice.name
        for choice in enum
    ]
    if all((isinstance(enum_value, str) for enum_value in enum_values)):
        return dict(
            type="string",
            enum=enum_values,
        )
    elif all((is_int(enum_value) for enum_value in enum_values)):
        return dict(
            type="integer",
            enum=enum_values,
        )
    else:
        raise Exception("Cannot infer enum type for field: {}".format(field.name))


def resolve_timestamp_field(field):
    """
    Timestamp fields should respect configuration.

    """
    if field.use_isoformat:
        return dict(
            type="string",
            format="date-time",
        )
    else:
        return dict(
            type="float",
            format="timestamp",
        )


def resolve_numeric_string_field(field):
    """
    Numeric fields should respect the `as_string` parameter.

    """
    if isinstance(field, fields.Decimal):
        return dict(
            type="string",
            format="decimal",
        )
    else:
        return dict(
            type="string",
        )


def resolve_default_for_field(field):
    """
    Field defaults should be the fallback.

    """
    try:
        field_type, field_format = FIELD_MAPPINGS[type(field)]
        if field_format:
            return dict(
                type=field_type,
                format=field_format,
            )
        elif field_type:
            return dict(
                type=field_type,
            )
        else:
            return dict()
    except KeyError:
        logger.exception("No mapped swagger type for marshmallow field: {}".format(
            field,
        ))
        raise


def build_parameter(field):
    """
    Build a parameter from a marshmallow field.

    See: https://github.com/marshmallow-code/apispec/blob/dev/apispec/ext/marshmallow/swagger.py#L81

    """
    if hasattr(field, SWAGGER_TYPE):
        parameter = resolve_tagged_field(field)
    elif getattr(field, "enum", None):
        parameter = resolve_enum_field(field)
    elif getattr(field, 'as_string', None):
        parameter = resolve_numeric_string_field(field)
    elif isinstance(field, TimestampField):
        parameter = resolve_timestamp_field(field)
    else:
        parameter = resolve_default_for_field(field)

    if field.metadata.get("description"):
        parameter["description"] = field.metadata["description"]

    if field.default:
        parameter["default"] = field.default

    # nested
    if isinstance(field, fields.Nested):
        parameter["$ref"] = "#/definitions/{}".format(type_name(name_for(field.schema)))

    # arrays
    if isinstance(field, fields.List):
        parameter["items"] = build_parameter(field.container)

    return parameter


def build_schema(marshmallow_schema):
    """
    Build JSON schema from a marshmallow schema.

    """
    fields = list(iter_fields(marshmallow_schema))
    required_fields = [
        field.dump_to or name
        for name, field in fields
        if field.required and not field.allow_none
    ]
    schema = {
        "type": "object",
        "properties": {
            field.dump_to or name: build_parameter(field)
            for name, field in fields
        }
    }
    if required_fields:
        schema["required"] = required_fields
    return schema


def iter_fields(marshmallow_schema):
    """
    Iterate through marshmallow schema fields.

    Generates: name, field pairs

    """
    for name in sorted(marshmallow_schema.fields.keys()):
        yield name, marshmallow_schema.fields[name]


def iter_schemas(marshmallow_schema):
    """
    Build zero or more JSON schemas for a marshmallow schema.

    Generates: name, schema pairs.

    """
    if not marshmallow_schema:
        return

    base_schema = build_schema(marshmallow_schema)
    base_schema_name = type_name(name_for(marshmallow_schema))
    yield base_schema_name, base_schema

    for name, field in iter_fields(marshmallow_schema):
        if isinstance(field, fields.Nested):
            nested_schema = build_schema(field.schema)
            nested_schema_name = type_name(name_for(field.schema))
            yield nested_schema_name, nested_schema
            for subname, subfield in iter_schemas(field.schema):
                yield subname, subfield
        if isinstance(field, fields.List) and isinstance(field.container, fields.Nested):
            nested_schema = build_schema(field.container.schema)
            nested_schema_name = type_name(name_for(field.container.schema))
            yield nested_schema_name, nested_schema
            for subname, subfield in iter_schemas(field.container.schema):
                yield subname, subfield



"""
Health check convention.

Reports service health and basic information from the "/api/health" endpoint,
using HTTP 200/503 status codes to indicate healthiness.

"""
from distutils.util import strtobool
from microcosm.api import defaults

from microcosm_flask.audit import skip_logging
from microcosm_flask.conventions.base import Convention
from microcosm_flask.conventions.build_info import BuildInfo
from microcosm_flask.conventions.encoding import make_response
from microcosm_flask.errors import extract_error_message
from microcosm_flask.namespaces import Namespace
from microcosm_flask.operations import Operation


class HealthResult:
    def __init__(self, error=None, result=None):
        self.error = error
        self.result = result or "ok"

    def __nonzero__(self):
        return self.error is None

    def __bool__(self):
        return self.error is None

    def __str__(self):
        return self.result if self.error is None else self.error

    def to_dict(self):
        return {
            "ok": bool(self),
            "message": str(self),
        }

    @classmethod
    def evaluate(cls, func, graph):
        try:
            result = func(graph)
            return cls(result=result)
        except Exception as error:
            return cls(error=extract_error_message(error))


class Health:
    """
    Wrapper around service health state.

    May contain zero or more "checks" which are just callables that take the
    current object graph as input.

    The overall health is OK if all checks are OK.

    """
    def __init__(self, graph, include_build_info=True):
        self.graph = graph
        self.name = graph.metadata.name
        if include_build_info:
            self.checks = dict(
                build_num=BuildInfo.check_build_num,
                sha1=BuildInfo.check_sha1,
            )
        else:
            self.checks = dict()

    def to_dict(self):
        """
        Encode the name, the status of all checks, and the current overall status.

        """
        # evaluate checks
        checks = {
            key: HealthResult.evaluate(func, self.graph)
            for key, func in self.checks.items()
        }
        dct = dict(
            # return the service name helps for routing debugging
            name=self.name,
            ok=all(checks.values()),
        )
        if checks:
            dct["checks"] = {
                key: checks[key].to_dict()
                for key in sorted(checks.keys())
            }
        return dct


class HealthConvention(Convention):

    def __init__(self, graph, include_build_info=False):
        super(HealthConvention, self).__init__(graph)
        self.health = Health(graph, include_build_info)

    def configure_retrieve(self, ns, definition):

        @self.add_route(ns.singleton_path, Operation.Retrieve, ns)
        @skip_logging
        def current_health():
            response_data = self.health.to_dict()
            status_code = 200 if response_data["ok"] else 503
            return make_response(response_data, status_code=status_code)


@defaults(
    include_build_info="true",
)
def configure_health(graph):
    """
    Configure the health endpoint.

    :returns: a handle to the `Health` object, allowing other components to
              manipulate health state.
    """
    ns = Namespace(
        subject=Health,
    )

    include_build_info = strtobool(graph.config.health_convention.include_build_info)
    convention = HealthConvention(graph, include_build_info)
    convention.configure(ns, retrieve=tuple())
    return convention.health
